/*
    Copyright  1995-2001, The AROS Development Team. All rights reserved.
    $Id: parseix.c,v 1.3 2016/07/25 21:30:54 itix Exp $

    Desc:
    Lang: English
*/

#ifndef  DEBUG
#define  DEBUG 0
#endif

#include "cxintern.h"

#include <aros/debug.h>

#include <libraries/commodities.h>
#include <proto/commodities.h>
#define __NOLIBBASE__ 1
#include <proto/utility.h>
#include <proto/keymap.h>
#undef __NOLIBBASE__
#include <devices/inputevent.h>
#include <devices/keymap.h>
#include <string.h>

#include "parse.h"

#define DEBUG_PARSEIX(x)	;
#define DEBUG_PMATCH(x)		;

static
BOOL pMatch(const pix_S *, CONST_STRPTR, LONG *, BOOL *, struct Library *CxBase);

/*****************************************************************************/

static inline
BOOL IsSeparator(UBYTE a)
{
	// 29-dec-2002 bugfix: '\0' is not a separator! - Piru
	//if (a == ' ' || a == '\n' || a == '\t' || a == ',' || a == '\0')
	if (a == ' ' || a == '\t' || a == ',' || a == '\n')
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

static inline
VOID GetNext(STRPTR *str)
{
	while (!(IsSeparator(**str)) && **str != '\0')
	{
		(*str)++;
	}

	while (IsSeparator(**str) && **str != '\0')
	{
		(*str)++;
	}
}

/*****************************************************************************

    NAME */

AROS_LH2(LONG, ParseIX,

         /*  SYNOPSIS */

         AROS_LHA(STRPTR, desc, A0),
         AROS_LHA(IX *  , ix  , A1),

         /*  LOCATION */

         struct Library *, CxBase, 22, Commodities)

/*  FUNCTION

    Fill in an InputXpression 'ix' according to the specifications given
    in the string pointed to by 'desc'.

    The string should be formatted according to:

    [class] {[-] (qualifier|synonym)} [[-] upstroke] [HighMap|ANSICode]

    For more information on this, consult "xxx/CxParse.doc".

    INPUTS

    desc  --  pointer to the string specifying the conditions and codes of
              the InputXpression.
    ix    --  pointer to an (uninitialized) InputXpression structure that
              will be filled according to 'desc'.

    RESULT

    0   --  Everything went OK.
    -1  --  Tokens after end
    -2  --  'desc' was NULL

    NOTES

    EXAMPLE

    BUGS

    SEE ALSO

    MatchIX(), <libraries/commodities.h>

    INTERNALS

    HISTORY

    10.05.97   SDuvan  implemented

******************************************************************************/

{
	AROS_LIBFUNC_INIT

	LONG val;
	BOOL dash, upstrdash;

	/* Set as standard if no class is specified in the description */
#if 0
	/* Caller is expected to set this! */
	ix->ix_Version   = IX_VERSION;
#endif
	ix->ix_Class     = IECLASS_RAWKEY;
	ix->ix_Code      = 0;
	ix->ix_CodeMask  = 0xFFFF;
	ix->ix_Qualifier = 0;
	ix->ix_QualMask  = IX_NORMALQUALS & ~(IEQUALIFIER_INTERRUPT | IEQUALIFIER_MULTIBROADCAST);
	ix->ix_QualSame  = 0;

	/* empty descriptions strings shouldnt generate IECLASS_RAWKEY code0 ... (cyfm 2003/01/26) */
	if (desc == NULL || *desc == '\0')
	{
		DEBUG_PARSEIX(dprintf("ParseIX: desc == NULL, return -2\n"));

		ix->ix_Code = 0xFFFF;

		return -2;
	}

	DEBUG_PARSEIX(dprintf("ParseIX: ix = 0x%lx, desc = \"%s\"\n", ix, desc));

	while (IsSeparator(*desc))
	{
		desc++;
	}

	dash = FALSE;

	if (pMatch(pix_Class, desc, &val, &dash, CxBase))
	{
		ix->ix_Class = val;
		GetNext(&desc);
	}

	for (;;)
	{
		dash = TRUE;

		if (pMatch(pix_IEvent, desc, &val, &dash, CxBase))
		{
			if (dash)
			{
				ix->ix_QualMask &= (ULONG) ~val;
			}
			else
			{
				ix->ix_Qualifier |= val;
			}

			GetNext(&desc);
		}
		else
		{
			dash = TRUE;

			if (pMatch(pix_Synonyms, desc, &val, &dash, CxBase))
			{
				ULONG x;

				ix->ix_QualSame |= val;

				switch (val)
				{
					case IXSYM_SHIFT:
						x = IXSYM_SHIFTMASK;
						break;

					case IXSYM_ALT:
						x = IXSYM_ALTMASK;
						break;

					case IXSYM_CAPS:
						x = IXSYM_CAPSMASK;
						break;

					default:
						x = 0;
						break;
				}

				if (dash)
				{
					ix->ix_QualMask &= (ULONG) ~x;
				}
				else
				{
					ix->ix_Qualifier |= x;
				}

				GetNext(&desc);
			}
			else
			{
				break;
			}
		}
	}

	upstrdash = TRUE;

	if (pMatch(pix_Upstroke, desc, &val, &upstrdash, CxBase))
	{
		if (upstrdash)
		{
			ix->ix_CodeMask &= ~IECODE_UP_PREFIX;
		}
		else
		{
			ix->ix_Code |= IECODE_UP_PREFIX;
		}

		GetNext(&desc);
	}

	dash = FALSE;

	if (ix->ix_Class == IECLASS_RAWMOUSE)
	{
		if (pMatch(pix_Rawmouse, desc, &val, &dash, CxBase))
		{
			DEBUG_PARSEIX(dprintf("pMatch successful -> ix->ix_Code |= val %ld\n",val);)

			ix->ix_Code |= val;

			GetNext(&desc);
		}
	}
	else if (ix->ix_Class == IECLASS_RAWKEY)
	{
		if (pMatch(pix_Rawkey, desc, &val, &dash, CxBase))
		{
			DEBUG_PARSEIX(dprintf("pMatch successful -> ix->ix_Code |= val %ld\n",val);)

			ix->ix_Code |= val;

			GetNext(&desc);
		}
		else if (*desc)
		{
			if (ix->ix_Qualifier & IEQUALIFIER_NUMERICPAD)
			{
				if (pMatch(pix_Numpad, desc, &val, &dash, CxBase))
				{
					DEBUG_PARSEIX(dprintf("pMatch successful -> ix->ix_Code |= val %ld\n",val);)

					ix->ix_Code |= val;

					GetNext(&desc);
				}
			}
			else
			{
				struct InputEvent event;

				DEBUG_PARSEIX(dprintf("might be literal key, try InvertKeyMap\n");)

				if (InvertKeyMap(*desc, &event, NULL))
				{
					ix->ix_Code      |= event.ie_Code;
					ix->ix_Qualifier |= event.ie_Qualifier;

					desc++;
					while (IsSeparator(*desc))
					{
						desc++;
					}
				}
			}
		}
	}

	DEBUG_PARSEIX(dprintf("ParseIX: Class 0x%lx Code 0x%lx CodeMask 0x%lx\n"
	                      "ParseIX: Qualifier 0x%lx QualMask 0x%lx QualSame 0x%lx\n",
	                      ix->ix_Class,
	                      ix->ix_Code,
	                      ix->ix_CodeMask,
	                      ix->ix_Qualifier,
	                      ix->ix_QualMask,
	                      ix->ix_QualSame));

	if (*desc == '\0')
	{
		DEBUG_PARSEIX(dprintf("ParseIX: return ok\n"));
		return 0;
	}
	else
	{
		DEBUG_PARSEIX(dprintf("ParseIX: tokens at end, desc 0x%lx *desc 0x%lx\n", desc, *desc));
		return -1;
	}

	AROS_LIBFUNC_EXIT
} /* ParseIX */


static
BOOL pMatch(const pix_S *words, CONST_STRPTR string, LONG *v, BOOL *dash,
            struct Library *CxBase)
{
	int i;

	DEBUG_PMATCH(dprintf("pMatch: words[0] = \"%s\" string \"%s\" dash %d\n",
	                     words[0].name, string, *dash));

	if (*dash)
	{
		if (*string == '-')
		{
			string++;
			//*dash = TRUE;
		}
		else
		{
			*dash = FALSE;
		}
	}

	for (i = 0; words[i].name; i++)
	{
		int wlen = strlen(words[i].name);

		if (Strnicmp(string, words[i].name, wlen) == 0 &&
		    (string[wlen] == '\0' || IsSeparator(string[wlen])))
		{
			*v = words[i].value;
			DEBUG_PMATCH(dprintf("pMatch: \"%s\" value 0x%lx\n", words[i].name, *v));
			return TRUE;
		}
	}

	DEBUG_PMATCH(dprintf("pMatch: not found\n"));
	return FALSE;
}
